// A constant which is used as the node attribute name for storing the previous hidden state
var CONTACT_SEARCH_ATTRIBUTE_NAME_LAST_STATE_HIDDEN = "data-last-state-is-hidden";

// Delimiter characters for detecting
var DELIMITER_CHAR_COMMA = ',';
var DELIMITER_CHAR_SEMICOLON = ';';

// The maximum height of the recipient row in pixels after which the lozenges will be collapsed
var COLLAPSE_LOZENGES_HEIGHT_THRESHOLD_PX = 150;

// A constant prefix which is part of every suggestion id
var CONTACT_SEARCH_ID_PREFIX_SUGGESTION = "ymail_contactSuggestion";

// A place holder token representing the suggestion title
var CONTACT_SEARCH_PLACEHOLDER_TITLE = "#CONTACTSEARCHTITLE#";

// A place holder token representing the suggestion subtitle
var CONTACT_SEARCH_PLACEHOLDER_SUBTITLE = "#CONTACTSEARCHSUBTITLE#";

var CONTACT_SEARCH_PERMISSION_PROMPT_POSITIVE_ACTION = "#PERMISSIONPOSITIVEACTION#";

var CONTACT_SEARCH_PERMISSION_PROMPT_NEGATIVE_ACTION = "#PERMISSIONNEGTIVEACTION#";

var CONTACT_SEARCH_PERMISSION_PROMPT_MESSAGE_ACTION = "#PERMISSIONMESSAGEACTION#";

var CONTACT_SUGGESTION_PROFILE_IMAGE = "<span class=\"left suggestion-profile-image\">" + "</span>";

// The contact suggestion HTML template
var CONTACT_SEARCH_TEMPLATE_SUGGESTION =
    CONTACT_SUGGESTION_PROFILE_IMAGE +
    "<span class=\"center prevent-text-selection suggestion-name\">" +
    "   <span class=\"prevent-text-selection suggestion-title\">" + CONTACT_SEARCH_PLACEHOLDER_TITLE + "</span>" +
    "   <br/>" +
    "   <span class=\"prevent-text-selection suggestion-subtitle\">" + CONTACT_SEARCH_PLACEHOLDER_SUBTITLE + "</span>" +
    "</span>";

var CONTACT_SEARCH_PERMISSION_PROMPT = "<span class=\"center prevent-text-selection permission-prompt\">" +
"   <span class=\"prevent-text-selection prompt-title\">" + CONTACT_SEARCH_PLACEHOLDER_TITLE + "</span>" +
"   <br/>" +
"   <span class=\"prevent-text-selection prompt-message\">" + CONTACT_SEARCH_PLACEHOLDER_SUBTITLE + "</span>" +
"</span>";

var PERMISSION_PROMPT_ACTION_BUTTONS =
"   <button class=\"prompt-cta permission-buttons allow-contacts\">" + CONTACT_SEARCH_PERMISSION_PROMPT_POSITIVE_ACTION + "</button>" +
"   <button class=\"prompt-cta permission-buttons not-now\">" + CONTACT_SEARCH_PERMISSION_PROMPT_NEGATIVE_ACTION + "</button>"

/**
 * An object which interacts with native code to provide type ahead search results
 */
var contactSearchController = {
    contactSearchManagers: [],
    isSearchModeOn: false,
    isEmptySearchSuggestionsEnabled: false,

    /**
     * A method which creates instances of ContactSearchManager and holds onto the created object references
     */
    create: function(recipientInputNode, searchSuggestionContainerNode, recipientRow, uniqueTag) {
        if (!this.contactSearchManagers[uniqueTag]) {
            this.contactSearchManagers[uniqueTag] = new ContactSearchManager(recipientInputNode, searchSuggestionContainerNode, recipientRow, uniqueTag);
        }
    },

    /**
     * Shows the contact search results for the given recipient field. The results are given in form of a JSON Array.
     * Stops the performance timer for the perfEvent (Online or Offline)
     */
    showResults: function(resultsArrayJsonString, uniqueTag, perfEvent, permissionPromptSuggestedObject) {
        this.contactSearchManagers[uniqueTag].showResults(resultsArrayJsonString, permissionPromptSuggestedObject);
        yMailBridge.stopAutoCompleteTimer(perfEvent);
    },

    /**
     * Destroys the search suggestion mode, by clearing all results
     */
    destroySearchMode: function(uniqueTag) {
        this.contactSearchManagers[uniqueTag].destroySearchMode();
    },

    close: function(srcInput) {
        for (const m of this.contactSearchManagers) {
            if (m && m.recipientInputNode && m.isShowing()) {
                if (m.recipientInputNode.id != srcInput) {
                    m.destroySearchMode();
                }
            }
        }
    },

    setEmptySearchSuggestionsFeatureFlag: function(enabled) {
        this.isEmptySearchSuggestionsEnabled = enabled;
    },
};

/**
 * Instances of this class manages type ahead search in a given recipient field
 */
function ContactSearchManager(recipientInputNode, searchSuggestionContainerNode, recipientRow, uniqueTag) {
    this.recipientInputNode = recipientInputNode;
    this.searchSuggestionContainerNode = searchSuggestionContainerNode;
    this.recipientRow = recipientRow;
    this.areLozengesCollapsed = false;
    this.uniqueTag = uniqueTag;

    this.recipientInputNode.addEventListener("input", this.handleInput.bind(this), false);
    this.recipientInputNode.addEventListener("focusin", this.focusHandler.bind(this), false);
}

/**
 * A method which handles user input and triggers the type ahead search
 */
ContactSearchManager.prototype.handleInput = function(event) {
    var typedText = event.target.value;
    yMailBridge.setRecipientTypedText(typedText, this.uniqueTag);

    // If empty search suggestions are on, always trigger the contact search
    // since this is the signal for the compose fragment to show results
    if (contactSearchController.isEmptySearchSuggestionsEnabled) {
        yMailBridge.triggerContactSearch(typedText, this.uniqueTag);
    } else {
        // If empty search is off, then only trigger the search for non-empty queries.
        if (typedText.length === 0) {
            // User has deleted all text in the input box, hence restore nodes to previous state
            this.destroySearchMode();
        } else {
            yMailBridge.triggerContactSearch(typedText, this.uniqueTag);
        }
    }

    // Check and turn typed text into lozenges
    var lastTypedChar = typedText[typedText.length - 1];
    if (lastTypedChar === DELIMITER_CHAR_COMMA || lastTypedChar === DELIMITER_CHAR_SEMICOLON) {
        yMailBridge.addTypedTextAsContactToRecipientField(typedText.substr(0, typedText.length - 1), this.uniqueTag);
    }
};

ContactSearchManager.prototype.focusHandler = function(event) {
    // Close other search controllers
    contactSearchController.close(event.target.id);
    if (contactSearchController.isEmptySearchSuggestionsEnabled) {
        // Start the search to display the empty query results
        yMailBridge.triggerContactSearch("", this.uniqueTag);
    }
};

/**
 * A method which handles click events of individual suggestion
 */
ContactSearchManager.prototype.handleSuggestionClick = function(event) {
    var suggestionNode = event.currentTarget;
    var email = suggestionNode.email;
    yMailBridge.addSuggestionToRecipientField(email, this.uniqueTag, PICK_SRC_SEARCH);
    this.recipientInputNode.value = "";

    if (contactSearchController.isEmptySearchSuggestionsEnabled) {
        // We should trigger a contact search here since we want to show suggestions in the empty state
        yMailBridge.triggerContactSearch("", this.uniqueTag);
    }

    this.recipientInputNode.focus();

    if (!contactSearchController.isEmptySearchSuggestionsEnabled) {
        this.destroySearchMode();
    }
};

/**
 * A method which handles click events of allow contacts button
 */
 ContactSearchManager.prototype.handleAllowContactsClick = function(event) {
    yMailBridge.allowContactsPermission();
};

/**
 * A method which handles click events of allow contacts in device settings
 */
 ContactSearchManager.prototype.handleAllowContactsInDeviceSettingClick = function(event) {
    yMailBridge.allowContactsPermissionInDeviceSetting();
};

/**
 * A method which handles click events of not now button
 */
 ContactSearchManager.prototype.handleNotNowClick = function(event) {
    yMailBridge.denyContactsPermission();
};

/**
 * A method which sets up the UI in search mode, by hiding/showing certain elements
 */
ContactSearchManager.prototype.setupSearchMode = function() {
    if (!contactSearchController.isSearchModeOn) {
        contactSearchController.isSearchModeOn = true;

        // If there are recipients already present as lozenges, the height of the recipient field might be really large
        // In such cases the suggestions list won't be visible any more, hence we collapse the lozenges in order to get more space
        if (this.recipientRow.clientHeight > COLLAPSE_LOZENGES_HEIGHT_THRESHOLD_PX) {
            this.areLozengesCollapsed = true;
            lozengeController.collapseLozenges(this.uniqueTag);
        }
        this.searchSuggestionContainerNode.classList.remove("hidden");
        this.searchSuggestionContainerNode.setAttribute("aria-expanded", "true");
    }
};

/**
 * A method which restores the UI from search mode into the previous state
 */
ContactSearchManager.prototype.destroySearchMode = function() {
    if (contactSearchController.isSearchModeOn) {
        contactSearchController.isSearchModeOn = false;

        this.searchSuggestionContainerNode.classList.add("hidden");
        this.searchSuggestionContainerNode.setAttribute("aria-expanded", "false");

        if (this.areLozengesCollapsed) {
            this.areLozengesCollapsed = false;
            lozengeController.expandLozenges(this.uniqueTag);
        }
    }

    yMailBridge.setSearchModeOff(this.recipientInputNode.value);
};

/**
 * Clears all the suggestions in the suggestions container node by remove all children
 */
ContactSearchManager.prototype.clearSuggestions = function() {
    // We need to retain the existence of the hidden class
    // If we don't do this, we may end up removing the hidden class at the end of this function
    var isHidden = this.searchSuggestionContainerNode.classList.contains("hidden");

    if (!isHidden) {
         this.searchSuggestionContainerNode.classList.add("hidden");
    }

    // Clear previous results
    while (this.searchSuggestionContainerNode.firstChild) {
        this.searchSuggestionContainerNode.removeChild(this.searchSuggestionContainerNode.firstChild);
    }

    if (!isHidden) {
        this.searchSuggestionContainerNode.classList.remove("hidden");
    }
};

ContactSearchManager.prototype.isShowing = function() {
    var isHidden = this.searchSuggestionContainerNode.classList.contains("hidden");
    return !isHidden;
}

/**
 * A method which accepts type ahead suggestion results in form a JSON Array String and shows them
 */
ContactSearchManager.prototype.showResults = function(resultsArrayJsonString, permissionPromptSuggestedObject) {
    var resultsArray,
        result,
        suggestionNode,
        i;

    if ((resultsArrayJsonString && this.recipientInputNode.value) || (contactSearchController.isEmptySearchSuggestionsEnabled && resultsArrayJsonString)) {
        try {
            resultsArray = JSON.parse(resultsArrayJsonString);
        } catch (err) {
            log.e(err);
        }

        this.clearSuggestions();

        if (resultsArray && resultsArray.length > 0 || permissionPromptSuggestedObject) {
            this.setupSearchMode();
            for (i = 0; i < resultsArray.length; i++) {
                result = resultsArray[i];
                suggestionNode = document.createElement("div");
                var title = decodeURIComponent(result.title);
                var id = decodeURIComponent(result.id);
                var subtitle = decodeURIComponent(result.subtitle);
                var image = decodeURIComponent(result.image);
                suggestionNode.setAttribute("role", "option");
                suggestionNode.setAttribute("id", utils.escapeHtml(CONTACT_SEARCH_ID_PREFIX_SUGGESTION + id));
                suggestionNode.setAttribute("aria-label", utils.escapeHtml(title + ". " + subtitle));
                suggestionNode.setAttribute("class", "header-row search-suggestion leftMargined");
                suggestionNode.innerHTML = CONTACT_SEARCH_TEMPLATE_SUGGESTION
                    .replace(CONTACT_SEARCH_PLACEHOLDER_TITLE, utils.escapeHtml(title))
                    .replace(CONTACT_SEARCH_PLACEHOLDER_SUBTITLE, utils.escapeHtml(subtitle));

                suggestionNode.getElementsByClassName("suggestion-profile-image")[0].classList.add("circle-image");

                profileImageController.load(image, suggestionNode.getElementsByClassName("suggestion-profile-image")[0], result.defaultImage);

                suggestionNode.email = id;
                suggestionNode.addEventListener("click", this.handleSuggestionClick.bind(this), false);
                this.searchSuggestionContainerNode.appendChild(suggestionNode);
            }
            if (permissionPromptSuggestedObject) {
                suggestionNode = document.createElement("div");
                var promptTitle = decodeURIComponent(permissionPromptSuggestedObject.promptTitle)
                suggestionNode.setAttribute("role", "option");
                suggestionNode.setAttribute("id", utils.escapeHtml(CONTACT_SEARCH_ID_PREFIX_SUGGESTION + permissionPromptSuggestedObject.id));
                suggestionNode.setAttribute("aria-label", utils.escapeHtml(promptTitle + ". " + permissionPromptSuggestedObject.subtitle));
                suggestionNode.setAttribute("class", "header-row search-suggestion leftMargined");
                var subtitle = permissionPromptSuggestedObject.promptMessage;
                if (permissionPromptSuggestedObject.promptMessageCta) {
                    subtitle = subtitle.replace("%1$s", "<span class=\"prompt-cta prevent-text-selection prompt-message-action\">" + permissionPromptSuggestedObject.promptMessageCta + "</span>");
                }
                suggestionNode.innerHTML = CONTACT_SEARCH_PERMISSION_PROMPT
                    .replace(CONTACT_SEARCH_PLACEHOLDER_TITLE, utils.escapeHtml(promptTitle))
                    .replace(CONTACT_SEARCH_PLACEHOLDER_SUBTITLE, subtitle);
                if (permissionPromptSuggestedObject.promptMessageCta) {
                    suggestionNode.querySelector(".prompt-message-action").addEventListener("click", this.handleAllowContactsInDeviceSettingClick.bind(this), false);
                }
                if (permissionPromptSuggestedObject.allowButton && permissionPromptSuggestedObject.notNowButton) {
                    suggestionNode.innerHTML += PERMISSION_PROMPT_ACTION_BUTTONS
                        .replace(CONTACT_SEARCH_PERMISSION_PROMPT_POSITIVE_ACTION, utils.escapeHtml(permissionPromptSuggestedObject.allowButton))
                        .replace(CONTACT_SEARCH_PERMISSION_PROMPT_NEGATIVE_ACTION, utils.escapeHtml(permissionPromptSuggestedObject.notNowButton));
                    suggestionNode.querySelector(".allow-contacts").addEventListener("click", this.handleAllowContactsClick.bind(this), false);
                    suggestionNode.querySelector(".not-now").addEventListener("click", this.handleNotNowClick.bind(this), false);
                }
                this.searchSuggestionContainerNode.appendChild(suggestionNode);
            }
        }
    }

    if (!resultsArray || (resultsArray.length === 0 && !permissionPromptSuggestedObject)) {
        this.destroySearchMode();
    }
};
